from __future__ import annotations

import abc
import inspect
import itertools
import json
import os
from collections import defaultdict
from collections.abc import Callable, Generator, Iterable
from dataclasses import dataclass
from datetime import date, datetime
from functools import cached_property
from typing import Any, ClassVar

from google.cloud import datastore
from pydantic import BaseModel

from gcp_pilot import exceptions

DEFAULT_NAMESPACE = os.environ.get("GCP_DATASTORE_NAMESPACE", default=None)
DEFAULT_PK_FIELD_NAME = "id"
DEFAULT_PK_FIELD_TYPE = int
MAX_ITEMS_PER_OPERATIONS = 500  # Datastore cannot write more than 500 items per call


def _chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


@dataclass
class DoesNotExist(Exception):
    cls: type[EmbeddedDocument]
    filters: dict


@dataclass
class MultipleObjectsFound(Exception):
    cls: type[EmbeddedDocument]
    filters: dict


def _starts_with_operator(lookup_fields, value) -> list[tuple[str, str, Any]]:
    field_name = ".".join(lookup_fields)
    return [
        (field_name, ">=", value),
        (field_name, "<=", f"{value}\ufffd"),
    ]


@dataclass
class Manager:
    LOOKUP_OPERATORS: ClassVar[dict[str, (str | Callable)]] = {
        "eq": "=",
        "gt": ">",
        "gte": ">=",
        "lt": "<",
        "lte": "<=",
        "in": "in",
        "startswith": _starts_with_operator,
    }
    doc_klass: type[EmbeddedDocument]

    @cached_property
    def client(self) -> datastore.Client:
        return datastore.Client(namespace=self.namespace)

    @property
    def namespace(self) -> str:
        return getattr(self.doc_klass.__config__, "namespace", ())

    @property
    def exclude_from_indexes(self):
        return getattr(self.doc_klass.__config__, "exclude_from_indexes", ())

    @property
    def kind(self) -> str:
        return "EmbeddedKind" if self.is_embedded else self.doc_klass.__name__

    @property
    def is_embedded(self) -> bool:
        return not issubclass(self.doc_klass, Document)

    @property
    def fields(self) -> Iterable[str]:
        return self.doc_klass.__fields__.keys()

    def build_key(self, pk: Any | None = None) -> datastore.Key:
        if self.is_embedded:
            return self.client.key(self.kind)

        if pk:
            typed_pk = DEFAULT_PK_FIELD_TYPE(pk)
            return self.client.key(self.kind, typed_pk)
        # If no primary key is provided, we let the server create a new ID
        return self.client.allocate_ids(self.client.key(self.kind), 1)[0]

    def _iterate(self, query, page_size: int = 10):
        cursor = None
        empty = False

        while not empty:
            query_iter = query.fetch(start_cursor=cursor, limit=page_size)

            page = next(query_iter.pages, [])
            yield from page
            next_cursor = query_iter.next_page_token
            empty = not bool(cursor) or cursor == next_cursor
            cursor = next_cursor

    def query(
        self,
        distinct_on: str | None = None,
        order_by: str | list[str] | None = None,
        page_size: int | None = None,
        **kwargs,
    ) -> datastore.query.Iterator:
        # base query
        query = self.client.query(kind=self.kind)
        if order_by:
            query.order = order_by
        if distinct_on:
            query.distinct_on = distinct_on

        # parse lookup args
        cross_filters = []
        for key, value in kwargs.items():
            all_filters = self._build_filter(key=key, value=value)
            for field_name, operator, field_value in all_filters:
                if operator == "in":
                    cross_filters.append((field_name, operator, field_value))
                else:
                    query.add_filter(field_name, operator, field_value)

        if not cross_filters:
            yield from self._iterate(query=query, page_size=page_size)
        else:
            # prepare combinations
            options = defaultdict(list)
            for name, operator, values in cross_filters:
                for value in values:
                    options[name].append((name, "=", value))

            combinations = itertools.product(*list(options.values()))

            found = set()
            for combination in combinations:
                current_query = query
                for field_name, operator, value in combination:
                    current_query = current_query.add_filter(field_name, operator, value)

                for item in self._iterate(query=current_query, page_size=page_size):
                    if item.id not in found:
                        yield item
                        found.add(item.id)

    def filter(self, **kwargs) -> Generator[Document, None, None]:
        for entity in self.query(**kwargs):
            yield self.doc_klass.from_entity(entity=entity)

    def get(self, **kwargs) -> Document:
        if DEFAULT_PK_FIELD_NAME in kwargs:
            pk = kwargs[DEFAULT_PK_FIELD_NAME]
            entity = self.client.get(key=self.build_key(pk=pk))
            if entity:
                return self.doc_klass.from_entity(entity=entity)
            raise DoesNotExist(self.doc_klass, pk)

        # Since we can't fetch directly from the key,
        # we filter and hope for just one object
        one_obj = None
        for obj in self.filter(**kwargs):
            if one_obj is not None:
                raise MultipleObjectsFound(self.doc_klass, filters=kwargs)
            one_obj = obj
        if not one_obj:
            raise DoesNotExist(self.doc_klass, filters=kwargs)
        return one_obj

    def create(self, **kwargs) -> Document:
        obj = self.doc_klass(**kwargs)
        entity = obj.to_entity()
        self.client.put(entity=entity)

        # if successfully saved, we assure the auto-generated ID is added to the final object
        if not obj.pk:
            setattr(obj, DEFAULT_PK_FIELD_NAME, entity.id)
        return obj

    def update(self, pk: str, **kwargs) -> Document:
        if kwargs:
            entity = self.client.get(key=self.build_key(pk=pk))
            # TODO: enable partial nested updates
            as_data = {
                key: value.to_dict() if isinstance(value, EmbeddedDocument) else value for key, value in kwargs.items()
            }
            entity.update(as_data)
            self.client.put(entity=entity)
        return self.get(id=pk)

    def delete(self, pk: str | None = None):
        if pk:
            self.client.delete(key=self.build_key(pk=pk))
        else:
            keys = [entity.key for entity in self.query()]
            for chunk in _chunks(keys, MAX_ITEMS_PER_OPERATIONS):
                self.client.delete_multi(keys=chunk)

    def _build_filter(self, key: str, value: Any) -> list[tuple[str, str, Any]]:
        operator = None

        *field_parts, lookup = key.split("__")
        if lookup:
            try:
                operator = self.LOOKUP_OPERATORS[lookup]

                if callable(operator):
                    field_name = ".".join(field_parts)
                    return operator(field_name, value)
            except KeyError:
                field_parts.append(lookup)

        if len(field_parts) > 1 and field_parts[0] not in self.fields:
            raise exceptions.ValidationError(
                f"{field_parts[0]} is not a valid field. Excepted one of {' | '.join(self.fields)}",
            )

        field_name = ".".join(field_parts)
        return [(field_name, (operator or "="), value)]


class EmbeddedDocument(BaseModel, abc.ABC):
    class Config:
        arbitrary_types_allowed = True
        use_enum_values = True
        keep_untouched = (cached_property,)

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        cls.documents = Manager(doc_klass=cls)

    @classmethod
    def from_dict(cls, **kwargs) -> EmbeddedDocument:
        return cls(**kwargs)

    def to_dict(self) -> dict:
        return self.dict()

    @classmethod
    def from_entity(cls, entity: datastore.Entity) -> EmbeddedDocument:
        data = dict(entity.items())
        data[DEFAULT_PK_FIELD_NAME] = entity.id
        return cls.from_dict(**data)

    def to_entity(self) -> datastore.Entity:
        exclude_from_indexes = self.documents.exclude_from_indexes
        if isinstance(self, Document):
            entity = datastore.Entity(
                key=self.documents.build_key(pk=self.pk),
                exclude_from_indexes=exclude_from_indexes,
            )
            if not self.pk:
                setattr(self, DEFAULT_PK_FIELD_NAME, entity.id)
        else:
            entity = datastore.Entity(
                key=self.documents.build_key(),
                exclude_from_indexes=exclude_from_indexes,
            )

        dict_obj = json.loads(self.json())

        data = {}
        for field_name, field_info in self.__fields__.items():
            if inspect.isclass(field_info.type_) and issubclass(field_info.type_, EmbeddedDocument):
                # recursively generate entities
                field_value = getattr(self, field_name)
                if isinstance(field_value, list | tuple):
                    value = [item.to_entity() for item in field_value]
                else:
                    value = field_value.to_entity() if field_value else None
            else:
                value = dict_obj[field_name]  # fetch json-friendly value
            data[field_name] = value

        entity.update(data)
        return entity


class Document(EmbeddedDocument, abc.ABC):
    id: DEFAULT_PK_FIELD_TYPE | None = None

    class Config(EmbeddedDocument.Config):
        exclude_from_indexes = ()
        namespace = DEFAULT_NAMESPACE
        json_encoders = {
            datetime: lambda v: v.isoformat(),
            date: lambda v: v.isoformat(),
        }

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.update_forward_refs()  # needed when base model has optional field o.O

    @property
    def pk(self):
        return getattr(self, "id", None)

    def save(self) -> Document:
        return self.documents.create(**self.dict())

    def delete(self) -> None:
        self.documents.delete(pk=self.pk)


__all__ = (
    "DoesNotExist",
    "MultipleObjectsFound",
    "EmbeddedDocument",
    "Document",
)



# More Information: https://cloud.google.com/healthcare-api/docs/reference/rest
import abc
import json
import logging
import math
from collections.abc import Callable, Generator
from dataclasses import dataclass
from typing import Any
from urllib.parse import parse_qsl, urlsplit

from fhir.resources.domainresource import DomainResource
from fhir.resources.identifier import Identifier

from gcp_pilot import exceptions
from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType, friendly_http_error

logger = logging.getLogger("gcp-pilot")


class HealthcareBase(DiscoveryMixin, GoogleCloudPilotAPI, abc.ABC):
    _scopes = ["https://www.googleapis.com/auth/cloud-healthcare"]

    def __init__(self, **kwargs):
        super().__init__(
            serviceName="healthcare",
            version="v1beta1",
            cache_discovery=False,
            **kwargs,
        )

    def _set_location(self, location: str | None = None) -> str:
        # Only some limited regions are available: https://cloud.google.com/healthcare-api/docs/concepts/regions
        valid_locations = [
            "northamerica-northeast1",
            "us-central1",
            "us-east1",
            "us-east4",
            "us-west1",
            "us-west2",
            "us-west3",
            "southamerica-east1",
            "europe-west2",
            "europe-west3",
            "europe-west4",
            "europe-west6",
            "asia-east1",
            "asia-east2",
            "asia-northeast1",
            "asia-northeast2",
            "asia-northeast3",
            "asia-south1",
            "asia-southeast1",
            "asia-southeast2",
            "australia-southeast1",
        ]
        location = super()._set_location(location=location)
        if location not in valid_locations:
            logger.warning(f"The location {location} is not available for Healthcare API yet. Using multi-region US.")
            location = "us"
        return location

    def _dataset_path(self, name: str, project_id: str | None = None, location: str | None = None) -> str:
        location_path = self._location_path(project_id=project_id, location=location)
        return f"{location_path}/datasets/{name}"

    def list_datasets(
        self,
        project_id: str | None = None,
        location: str | None = None,
    ) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._location_path(project_id=project_id, location=location),
        )

        yield from self._list(
            method=self.client.projects().locations().datasets().list,
            result_key="datasets",
            params=params,
        )

    def get_dataset(self, name: str, project_id: str | None = None, location: str | None = None) -> ResourceType:
        name = self._dataset_path(name=name, project_id=project_id, location=location)

        return self._execute(
            method=self.client.projects().locations().datasets().get,
            name=name,
        )

    def delete_dataset(self, name: str, project_id: str | None = None, location: str | None = None) -> ResourceType:
        name = self._dataset_path(name=name, project_id=project_id, location=location)

        return self._execute(
            method=self.client.projects().locations().datasets().delete,
            name=name,
        )

    def create_dataset(
        self,
        name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        parent = self._location_path(project_id=project_id, location=location)

        body = {}

        return self._execute(
            method=self.client.projects().locations().datasets().create,
            parent=parent,
            datasetId=name,
            body=body,
        )


def as_json(resource: DomainResource) -> dict[str, Any]:
    return json.loads(resource.json())


@dataclass
class FHIRResultSet:
    method: Callable
    url: str
    resource_class: type[DomainResource]
    query: dict[str, Any] | None = None
    order_by: str | None = None
    limit: int | None = None
    cursor: str | None = None
    _response: dict | None = None

    @property
    def response(self):
        if not self._response:
            self._request()
        return self._response

    @response.setter
    def response(self, value):
        self._response = value

    @friendly_http_error
    def _request(self):
        kwargs = {
            "url": self.url,
            "headers": {"Content-Type": "application/fhir+json;charset=utf-8"},
            "params": self.query or {},
        }
        if self.order_by:
            kwargs["params"]["_sort"] = self.order_by
        if self.limit:
            kwargs["params"]["_count"] = self.limit
        if self.cursor:
            kwargs["params"]["_page_token"] = self.cursor

        request = self.method(**kwargs)
        request.raise_for_status()
        self.response = request.json()

    def __iter__(self) -> Generator[dict, None, None]:
        while True:
            yield from self.get_page_resources()
            self.cursor = self.next_cursor
            if not self.cursor:
                break
            self._request()

    def first(self) -> dict:
        try:
            return next(iter(self))
        except StopIteration as exc:
            raise exceptions.NotFound() from exc

    def get_page_resources(self) -> Generator[DomainResource, None, None]:
        if self.total == 0:
            return

        for entry in self.response["entry"]:
            yield self.resource_class(**entry["resource"])

    @property
    def next_cursor(self) -> str | None:
        for link in self.response["link"]:
            if link["relation"] == "next":
                query_params = dict(parse_qsl(urlsplit(link["url"]).query))
                return query_params.get("_page_token")
        return None

    @property
    def previous_cursor(self) -> str | None:
        for link in self.response["link"]:
            if link["relation"] == "previous":
                return link["url"]
        return None

    @property
    def total(self) -> int:
        return self.response["total"]

    def __len__(self) -> int:
        return self.total

    @property
    def num_pages(self) -> int:
        return int(math.ceil(self.total / self.limit))


class HealthcareFHIR(HealthcareBase):
    def _store_path(
        self,
        name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> str:
        dataset_path = self._dataset_path(name=dataset_name, project_id=project_id, location=location)
        return f"{dataset_path}/fhirStores/{name}"

    def _resource_path(
        self,
        resource_type: str,
        resource_id: str,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> str:
        store_path = self._store_path(
            name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )
        return f"{store_path}/fhir/{resource_type}/{resource_id}"

    def _operation_path(
        self,
        operation_id: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> str:
        dataset_path = self._dataset_path(name=dataset_name, project_id=project_id, location=location)
        return f"{dataset_path}/operations/{operation_id}"

    def list_stores(
        self,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._dataset_path(name=dataset_name, project_id=project_id, location=location),
        )

        yield from self._list(
            method=self.client.projects().locations().datasets().fhirStores().list,
            result_key="fhirStores",
            params=params,
        )

    def get_store(
        self,
        name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._store_path(name=name, dataset_name=dataset_name, project_id=project_id, location=location)

        return self._execute(
            method=self.client.projects().locations().datasets().fhirStores().get,
            name=name,
        )

    def delete_store(
        self,
        name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._store_path(name=name, dataset_name=dataset_name, project_id=project_id, location=location)

        return self._execute(
            method=self.client.projects().locations().datasets().fhirStores().delete,
            name=name,
        )

    def list_resources(
        self,
        store_name: str,
        dataset_name: str,
        resource_class: type[DomainResource],
        project_id: str | None = None,
        location: str | None = None,
        query: dict[str, Any] | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> FHIRResultSet:
        parent = self._store_path(
            name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )
        resource_type = resource_class.get_resource_type()
        url = f"{self._base_url}/{parent}/fhir/{resource_type}/_search"

        return FHIRResultSet(
            method=self._session.post,
            url=url,
            resource_class=resource_class,
            query=query,
            limit=limit,
            cursor=cursor,
        )

    def create_store(
        self,
        name: str,
        dataset_name: str,
        labels: dict[str, str] | None = None,
        enable_upsert: bool = True,
        notify_pubsub_topic: str | None = None,
        export_to_bigquery_dataset: str | None = None,
        project_id: str | None = None,
        location: str | None = None,
        version: str = "R4",
    ) -> ResourceType:
        parent = self._dataset_path(name=dataset_name, project_id=project_id, location=location)

        body = {
            "name": name,
            "enableUpdateCreate": enable_upsert,
            "labels": labels,
            "version": version,
        }

        if notify_pubsub_topic:
            if "/" not in notify_pubsub_topic:
                notify_pubsub_topic = f"projects/{(project_id or self.project_id)}/topics/{notify_pubsub_topic}"
            body["notification_config"] = {
                "pubsubTopic": notify_pubsub_topic,
            }

        if export_to_bigquery_dataset:
            if "." not in export_to_bigquery_dataset:
                export_to_bigquery_dataset = f"{(project_id or self.project_id)}.{export_to_bigquery_dataset}"

            body["streamConfigs"] = {
                "resourceTypes": [],  # empty means all
                "bigqueryDestination": {
                    "datasetUri": f"bq://{export_to_bigquery_dataset}",
                    "schemaConfig": {
                        "schemaType": "ANALYTICS_V2",
                        "recursiveStructureDepth": 5,  # max depth
                    },
                    "writeDisposition": "WRITE_TRUNCATE",
                },
            }

        return self._execute(
            method=self.client.projects().locations().datasets().fhirStores().create,
            parent=parent,
            body=body,
            fhirStoreId=name,
        )

    def get_resource(
        self,
        resource_class: type[DomainResource],
        resource_id: str,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> DomainResource:
        resource_type = resource_class.get_resource_type()
        name = self._resource_path(
            resource_type=resource_type,
            resource_id=resource_id,
            store_name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )

        data = self._execute(
            method=self.client.projects().locations().datasets().fhirStores().fhir().read,
            name=name,
        )
        return resource_class(**data)

    def delete_resource(
        self,
        resource_class: type[DomainResource],
        resource_id: str,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        resource_type = resource_class.get_resource_type()
        name = self._resource_path(
            resource_type=resource_type,
            resource_id=resource_id,
            store_name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )

        data = self._execute(
            method=self.client.projects().locations().datasets().fhirStores().fhir().delete,
            name=name,
        )
        return data

    def create_resource(
        self,
        resource: DomainResource,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> DomainResource:
        parent = self._store_path(name=store_name, dataset_name=dataset_name, project_id=project_id, location=location)
        data = self._execute(
            method=self.client.projects().locations().datasets().fhirStores().fhir().create,
            parent=parent,
            type=resource.get_resource_type(),
            body=as_json(resource),
        )
        return resource.__class__(**data)

    def update_resource(
        self,
        resource: DomainResource,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> DomainResource:
        name = self._resource_path(
            resource_type=resource.get_resource_type(),
            resource_id=resource.id,
            store_name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )

        data = self._execute(
            method=self.client.projects().locations().datasets().fhirStores().fhir().update,
            name=name,
            body=as_json(resource),
        )
        return resource.__class__(**data)

    def patch_update(
        self,
        resource_class: type[DomainResource],
        data: dict,
        resource_id: str,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> DomainResource:
        name = self._resource_path(
            resource_type=resource_class.get_resource_type(),
            resource_id=resource_id,
            store_name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )
        body = [{"op": "replace", "path": f"/{key}", "value": value} for key, value in data.items()]
        data = self._execute(
            method=self.client.projects().locations().datasets().fhirStores().fhir().patch,
            method_http_headers={"Content-Type": "application/json-patch+json"},
            name=name,
            body=body,
        )
        return resource_class(**data)

    def create_or_update_resource(
        self,
        resource: DomainResource,
        store_name: str,
        dataset_name: str,
        query: dict[str, Any] | None = None,
        project_id: str | None = None,
        location: str | None = None,
    ) -> DomainResource:
        if not query:
            identifiers: list[Identifier] = getattr(resource, "identifier")
            if not identifiers:
                raise exceptions.ValidationError("Either `query` or identifiers must be provided to create-or-update")

            query_values = []
            for identifier in identifiers:
                query_values.append(identifier.system)
                query_values.append(identifier.value)
            query = {"identifier": "|".join(query_values)}

        parent = self._store_path(
            name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )
        url = f"{self._base_url}/{parent}/fhir/{resource.get_resource_type()}"

        params = dict(
            url=url,
            params=query,
            headers={"Content-Type": "application/fhir+json;charset=utf-8"},
            json=as_json(resource),
        )

        try:
            data = self._execute(method=self._session.put, **params)
        except exceptions.OperationError as exc:
            if exc.errors[0]["code"] == "conflict":
                raise exceptions.FailedPrecondition(
                    f"{resource.get_resource_type()} with query {query} matches multiple resources at {parent}",
                ) from exc
            raise

        return resource.__class__(**data)

    def export_resources(
        self,
        gcs_path: str,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
        resource_names: list[str] | None = None,
    ) -> dict[str, str]:
        name = self._store_path(name=store_name, dataset_name=dataset_name, project_id=project_id, location=location)
        body = {"gcsDestination": {"uriPrefix": f"gs://{gcs_path}"}}
        if resource_names:
            body["_type"] = ",".join(resource_names)
        data = self._execute(
            method=self.client.projects().locations().datasets().fhirStores().export,
            name=name,
            body=body,
        )

        _, project_id, _, location, _, dataset_name, _, operation_id = data["name"].split("/")
        return {
            "project_id": project_id,
            "location": location,
            "dataset_name": dataset_name,
            "operation_id": operation_id,
        }

    def get_operation(
        self,
        operation_id: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._operation_path(
            operation_id=operation_id,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )
        return self._execute(
            method=self.client.projects().locations().datasets().operations().get,
            name=name,
        )

    def import_resources(
        self,
        gcs_path: str,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        parent = self._store_path(name=store_name, dataset_name=dataset_name, project_id=project_id, location=location)
        body = {
            "contentStructure": "CONTENT_STRUCTURE_UNSPECIFIED",
            "gcsSource": {"uri": f"gs://{gcs_path}"},
        }
        return self._execute(
            method=self.client.projects().locations().datasets().fhirStores().import_,
            name=parent,
            body=body,
        )

    def get_resource_history(
        self,
        resource_class: type[DomainResource],
        resource_id: str,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> tuple[str, ResourceType | None]:
        name = self._resource_path(
            resource_type=resource_class.get_resource_type(),
            resource_id=resource_id,
            store_name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )

        entries = self._list(
            method=self.client.projects().locations().datasets().fhirStores().fhir().history,
            params=dict(
                name=name,
            ),
            result_key="entry",
        )

        for entry in entries:
            method = entry["request"]["method"]
            version = resource_class(**entry["resource"]) if "resource" in entry else None
            yield method, version

    def validate_resource(
        self,
        resource: DomainResource,
        store_name: str,
        dataset_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ):
        parent = self._store_path(
            name=store_name,
            dataset_name=dataset_name,
            project_id=project_id,
            location=location,
        )

        try:
            self._execute(
                method=self.client.projects().locations().datasets().fhirStores().fhir().Resource_validate,
                parent=parent,
                type=resource.get_resource_type(),
                body=as_json(resource),
            )
        except exceptions.OperationError as exc:
            errors = [{"fields": error["expression"], "detail": error["diagnostics"]} for error in exc.errors]
            raise exceptions.ValidationError(errors) from exc



# More Information: <https://cloud.google.com/logging/docs/reference/libraries#client-libraries-install-python>
from datetime import UTC, datetime
from logging import INFO

from google.cloud import logging
from google.cloud.logging_v2 import Resource
from google.logging.type.http_request_pb2 import HttpRequest
from google.logging.type.log_severity_pb2 import LogSeverity

from gcp_pilot.base import GoogleCloudPilotAPI


class CloudLogging(GoogleCloudPilotAPI):
    _client_class = logging.Client

    def _get_client_extra_kwargs(self):
        return {"project": self.project_id}

    def enable(self, log_level=INFO):
        self.client.setup_logging(log_level=log_level)

    @property
    def handler(self):
        return self.client.get_default_handler()

    def log_struct(
        self,
        logger_name: str,
        message: str | dict,
        severity: LogSeverity,
        timestamp: datetime | None = None,
        http_request: HttpRequest | None = None,
        labels: dict | None = None,
        span_id: str | None = None,
        trace: str | None = None,
        resource: Resource | None = None,
    ):
        logger = self.client.logger(name=logger_name)
        logger.log_struct(
            message,
            timestamp=timestamp or datetime.now(tz=UTC),
            labels=labels,
            resource=resource,
            severity=severity,
            trace=trace,
            span_id=span_id,
            http_request=http_request,
        )


__all__ = ("CloudLogging",)



# More Information: https://googleapis.dev/python/storage/latest/index.html
import io
from collections.abc import Generator
from pathlib import Path

import requests
from google.cloud import storage
from google.cloud.exceptions import Conflict
from google.cloud.storage import Blob, Bucket

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI


class CloudStorage(GoogleCloudPilotAPI):
    _client_class = storage.Client

    def create_bucket(
        self,
        name: str,
        region: str | None = None,
        project_id: str | None = None,
        exists_ok: bool = True,
    ) -> Bucket:
        bucket = self.client.bucket(name)
        try:
            return self.client.create_bucket(
                bucket_or_name=bucket,
                location=region,
                project=project_id or self.project_id,
            )
        except Conflict:
            if not exists_ok:
                raise
            return self.check_bucket(name=name)

    def check_bucket(self, name: str) -> Bucket:
        return self.client.get_bucket(bucket_or_name=name)

    def upload(
        self,
        source_file,
        bucket_name: str,
        target_file_name: str | None = None,
        chunk_size: int | None = None,
        is_public: bool = False,
        content_type: str | None = None,
    ) -> Blob:
        target_bucket = self.check_bucket(name=bucket_name)

        target_file_name = target_file_name or str(source_file).rsplit("/", maxsplit=1)[-1]
        blob = target_bucket.blob(target_file_name, chunk_size=chunk_size)

        if isinstance(source_file, str):
            if source_file.startswith("http"):
                file_obj = self._download(url=source_file)
                blob.upload_from_file(file_obj, content_type=content_type)
            elif Path(source_file).exists():
                blob.upload_from_filename(source_file, content_type=content_type)
            else:
                content = io.StringIO(source_file)
                blob.upload_from_file(content, content_type=content_type)
        elif isinstance(source_file, bytes):
            blob.upload_from_string(data=source_file, content_type=content_type)
        else:
            blob.upload_from_file(file_obj=source_file, content_type=content_type)

        if is_public:
            blob.make_public()

        return blob

    def copy(
        self,
        source_file_name,
        source_bucket_name: str,
        target_bucket_name: str,
        target_file_name: str | None = None,
        project_id: str | None = None,
        region: str | None = None,
    ) -> Blob:
        source_bucket = self.check_bucket(name=source_bucket_name)
        source_blob = source_bucket.blob(source_file_name)

        target_bucket = self.create_bucket(name=target_bucket_name, region=region, project_id=project_id)
        target_file_name = target_file_name or str(source_file_name).rsplit("/", maxsplit=1)[-1]

        obj = source_bucket.copy_blob(source_blob, target_bucket, target_file_name)
        return obj

    def move(
        self,
        source_file_name: str,
        source_bucket_name: str,
        target_bucket_name: str,
        target_file_name: str | None = None,
        project_id: str | None = None,
        region: str | None = None,
    ) -> Blob:
        data = self.copy(
            source_file_name=source_file_name,
            source_bucket_name=source_bucket_name,
            target_file_name=target_file_name,
            target_bucket_name=target_bucket_name,
            project_id=project_id,
            region=region,
        )
        self.delete(file_name=source_file_name, bucket_name=source_bucket_name)
        return data

    def delete(self, file_name: str, bucket_name: str | None = None) -> None:
        bucket = self.check_bucket(name=bucket_name)
        blob = bucket.blob(file_name)
        return blob.delete()

    def list_files(self, bucket_name: str, prefix: str | None = None) -> Generator[Blob, None, None]:
        blobs = self.client.list_blobs(
            bucket_name,
            prefix=prefix,
        )
        yield from blobs

    def get_file(self, uri: str) -> Blob:
        if not uri.startswith("gs://"):
            raise exceptions.ValidationError("GCS file must start with gs://")

        bucket_name, file_path = uri.removeprefix("gs://").split("/", 1)
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(file_path)

        return blob

    def _download(self, url: str) -> io.BytesIO:
        response = requests.get(url, stream=True, timeout=15)
        file = io.BytesIO()
        file.write(response.content)
        file.seek(0)
        return file

    def get_uri(self, blob: Blob) -> str:
        return f"gs://{blob.bucket.name}/{blob.name}"


__all__ = ("CloudStorage",)



# More Information: <https://cloud.google.com/resource-manager/reference/rest>
import logging
from collections.abc import Generator
from pathlib import Path

from gcp_pilot import exceptions
from gcp_pilot.base import AccountManagerMixin, DiscoveryMixin, GoogleCloudPilotAPI, PolicyType, ResourceType

logger = logging.getLogger("gcp-pilot")


class ResourceManager(AccountManagerMixin, DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName="cloudresourcemanager",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def get_policy(self, project_id: str | None = None, version: int = 1) -> PolicyType:
        return self._execute(
            method=self.client.projects().getIamPolicy,
            resource=project_id or self.project_id,
            body={"options": {"requestedPolicyVersion": version}},
        )

    def set_policy(self, policy: PolicyType, project_id: str | None = None) -> PolicyType:
        if not policy["bindings"]:
            raise exceptions.NotAllowed("Too dangerous to set policy with empty bindings")

        return self._execute(
            method=self.client.projects().setIamPolicy,
            resource=project_id or self.project_id,
            body={"policy": policy, "updateMask": "bindings"},
        )

    def add_member(self, email: str, role: str, project_id: str | None = None) -> PolicyType:
        policy = self.get_policy(project_id=project_id)
        changed_policy = self._bind_email_to_policy(email=email, role=role, policy=policy)
        return self.set_policy(policy=changed_policy, project_id=project_id)

    def remove_member(self, email: str, role: str, project_id: str | None = None) -> PolicyType:
        policy = self.get_policy(project_id=project_id)
        changed_policy = self._unbind_email_from_policy(email=email, role=role, policy=policy)
        return self.set_policy(policy=changed_policy, project_id=project_id)

    def get_project(self, project_id: str) -> ResourceType:
        return self._execute(
            method=self.client.projects().get,
            projectId=project_id or self.project_id,
        )

    def allow_impersonation(self, email: str, project_id: str | None = None) -> PolicyType:
        return self.add_member(
            email=email,
            role="roles/iam.serviceAccountTokenCreator",
            project_id=project_id,
        )


class ServiceAgent:
    agents = {
        "AI Platform Custom Code Service Agent": ("gcp-sa-aiplatform-cc.iam.gserviceaccount.com", None),
        "AI Platform Service Agent": ("gcp-sa-aiplatform.iam.gserviceaccount.com", None),
        "ASM Mesh Data Plane Service Account": (
            "gcp-sa-meshdataplane.iam.gserviceaccount.com",
            "roles/meshdataplane.serviceAgent",
        ),
        "Anthos Audit Service Account": ("gcp-sa-anthosaudit.iam.gserviceaccount.com", None),
        "Anthos Config Management Service Account": (
            "gcp-sa-anthosconfigmanagement.iam.gserviceaccount.com",
            "roles/anthosconfigmanagement.serviceAgent",
        ),
        "Anthos Identity Service Account": (
            "gcp-sa-anthosidentityservice.iam.gserviceaccount.com",
            "roles/anthosidentityservice.serviceAgent",
        ),
        "Anthos Service Account": ("gcp-sa-anthos.iam.gserviceaccount.com", "roles/anthos.serviceAgent"),
        "Anthos Service Mesh Service Account": ("gcp-sa-servicemesh.iam.gserviceaccount.com", None),
        "Apigee Service Agent": ("gcp-sa-apigee.iam.gserviceaccount.com", "roles/apigee.serviceAgent"),
        "App Engine Flexible Environment Service Agent": (
            "gae-api-prod.google.com.iam.gserviceaccount.com",
            "roles/appengineflex.serviceAgent",
        ),
        "Artifact Registry Service Agent": (
            "gcp-sa-artifactregistry.iam.gserviceaccount.com",
            "roles/artifactregistry.serviceAgent",
        ),
        "AssuredWorkloads Service Account": ("gcp-sa-assuredworkloads.iam.gserviceaccount.com", None),
        "AutoML Recommendations Service Account": (
            "gcp-sa-recommendationengine.iam.gserviceaccount.com",
            "roles/automlrecommendations.serviceAgent",
        ),
        "AutoML Service Agent": ("gcp-sa-automl.iam.gserviceaccount.com", "roles/automl.serviceAgent"),
        "BigQuery Connection Service Agent": (
            "gcp-sa-bigqueryconnection.iam.gserviceaccount.com",
            "roles/bigqueryconnection.serviceAgent",
        ),
        "BigQuery Data Transfer Service Agent": (
            "gcp-sa-bigquerydatatransfer.iam.gserviceaccount.com",
            "roles/bigquerydatatransfer.serviceAgent",
        ),
        "BigQuery Omni Service Agent": ("gcp-sa-prod-bigqueryomni.iam.gserviceaccount.com", None),
        "Binary Authorization Service Agent": (
            "gcp-sa-binaryauthorization.iam.gserviceaccount.com",
            "roles/binaryauthorization.serviceAgent",
        ),
        "Cloud AI Platform Notebooks Service Account": (
            "gcp-sa-notebooks.iam.gserviceaccount.com",
            "roles/notebooks.serviceAgent",
        ),
        "Cloud API Gateway Management Plane Service Account": ("gcp-sa-apigateway-mgmt.iam.gserviceaccount.com", None),
        "Cloud API Gateway Service Account": ("gcp-sa-apigateway.iam.gserviceaccount.com", None),
        "Cloud Asset Service Agent": ("gcp-sa-cloudasset.iam.gserviceaccount.com", "roles/cloudasset.serviceAgent"),
        "Cloud Bigtable Service Agent": ("gcp-sa-bigtable.iam.gserviceaccount.com", None),
        "Cloud Build Service Agent": ("gcp-sa-cloudbuild.iam.gserviceaccount.com", "roles/cloudbuild.serviceAgent"),
        "Cloud Composer Service Agent": (
            "cloudcomposer-accounts.iam.gserviceaccount.com",
            "roles/composer.serviceAgent",
        ),
        "Cloud Data Fusion Service Account": (
            "gcp-sa-datafusion.iam.gserviceaccount.com",
            "roles/datafusion.serviceAgent",
        ),
        "Cloud Database Migration Service Account": ("gcp-sa-datamigration.iam.gserviceaccount.com", None),
        "Cloud Dataflow Service Account": (
            "dataflow-service-producer-prod.iam.gserviceaccount.com",
            "roles/dataflow.serviceAgent",
        ),
        "Cloud Endpoints Service Agent": ("gcp-sa-endpoints.iam.gserviceaccount.com", "roles/endpoints.serviceAgent"),
        "Cloud File Storage Service Account": ("cloud-filer.iam.gserviceaccount.com", "roles/file.serviceAgent"),
        "Cloud Healthcare Service Agent": (
            "gcp-sa-healthcare.iam.gserviceaccount.com",
            "roles/healthcare.serviceAgent",
        ),
        "Cloud IoT Core Service Agent": ("gcp-sa-cloudiot.iam.gserviceaccount.com", "roles/cloudiot.serviceAgent"),
        "Cloud KMS Service Agent": ("gcp-sa-cloudkms.iam.gserviceaccount.com", "roles/cloudkms.serviceAgent"),
        "Cloud Life Sciences Service Agent": (
            "gcp-sa-lifesciences.iam.gserviceaccount.com",
            "roles/lifesciences.serviceAgent",
        ),
        "Cloud Logging Service Account": ("gcp-sa-logging.iam.gserviceaccount.com", None),
        "Cloud Managed Identities Service Agent": (
            "gcp-sa-mi.iam.gserviceaccount.com",
            "roles/managedidentities.serviceAgent",
        ),
        "Cloud Memorystore Memcache Service Agent": (
            "cloud-memcache-sa.iam.gserviceaccount.com",
            "roles/memcache.serviceAgent",
        ),
        "Cloud Memorystore Redis Service Agent": ("cloud-redis.iam.gserviceaccount.com", "roles/redis.serviceAgent"),
        "Cloud Network Management Service Account": (
            "gcp-sa-networkmanagement.iam.gserviceaccount.com",
            "roles/networkmanagement.serviceAgent",
        ),
        "Cloud Pub/Sub Service Account": ("gcp-sa-pubsub.iam.gserviceaccount.com", None),
        "Cloud SQL Service Account": ("gcp-sa-cloud-sql.iam.gserviceaccount.com", "roles/cloudsql.serviceAgent"),
        "Cloud Scheduler Service Account": (
            "gcp-sa-cloudscheduler.iam.gserviceaccount.com",
            "roles/cloudscheduler.serviceAgent",
        ),
        "Cloud Security Command Center Notification Service Account": (
            "gcp-sa-scc-notification.iam.gserviceaccount.com",
            "roles/securitycenter.notificationServiceAgent",
        ),
        "Cloud Source Repositories Service Agent": (
            "sourcerepo-service-accounts.iam.gserviceaccount.com",
            "roles/sourcerepo.serviceAgent",
        ),
        "Cloud Spanner Production Service Account": ("gcp-sa-spanner.iam.gserviceaccount.com", None),
        "Cloud Storage for Firebase Service Agent": (
            "gcp-sa-firebasestorage.iam.gserviceaccount.com",
            "roles/firebasestorage.serviceAgent",
        ),
        "Cloud Tasks Service Account": ("gcp-sa-cloudtasks.iam.gserviceaccount.com", "roles/cloudtasks.serviceAgent"),
        "Cloud Trace Service Account": ("gcp-sa-cloud-trace.iam.gserviceaccount.com", None),
        "Cloud Translation Service Agent": (
            "gcp-sa-translation.iam.gserviceaccount.com",
            "roles/cloudtranslate.serviceAgent",
        ),
        "Cloud VM Migration Service Account": ("gcp-sa-vmmigration.iam.gserviceaccount.com", None),
        "Cloud Web Security Scanner Service Agent": (
            "gcp-sa-websecurityscanner.iam.gserviceaccount.com",
            "roles/websecurityscanner.serviceAgent",
        ),
        "Cloud Workflows Service Agent": ("gcp-sa-workflows.iam.gserviceaccount.com", "roles/workflows.serviceAgent"),
        "Compute Engine Service Agent": ("compute-system.iam.gserviceaccount.com", "roles/compute.serviceAgent"),
        "Compute Scanning Service Agent": (
            "gcp-sa-computescanning.iam.gserviceaccount.com",
            "roles/computescanning.serviceAgent",
        ),
        "Container Analysis Service Agent": (
            "container-analysis.iam.gserviceaccount.com",
            "roles/containeranalysis.ServiceAgent",
        ),
        "Container Scanning Service Agent": (
            "gcp-sa-containerscanning.iam.gserviceaccount.com",
            "roles/containerscanning.ServiceAgent",
        ),
        "Container Threat Detection Service Agent": (
            "gcp-sa-ktd-control.iam.gserviceaccount.com",
            "roles/containerthreatdetection.serviceAgent",
        ),
        "Data Labeling Service Account": (
            "gcp-sa-datalabeling.iam.gserviceaccount.com",
            "roles/datalabeling.serviceAgent",
        ),
        "Data Studio Service Account": ("gcp-sa-datastudio.iam.gserviceaccount.com", "roles/datastudio.serviceAgent"),
        "Dataproc Metastore Service Account": ("gcp-sa-metastore.iam.gserviceaccount.com", None),
        "Dialogflow Service Agent": ("gcp-sa-dialogflow.iam.gserviceaccount.com", "roles/dialogflow.serviceAgent"),
        "DocumentAI Core Service Agent": (
            "gcp-sa-prod-dai-core.iam.gserviceaccount.com",
            "roles/documentaicore.serviceAgent",
        ),
        "Endpoints Consumer Portal Service Agent": (
            "endpoints-portal.iam.gserviceaccount.com",
            "roles/endpointsportal.serviceAgent",
        ),
        "Eventarc Service Agent": ("gcp-sa-eventarc.iam.gserviceaccount.com", "roles/eventarc.serviceAgent"),
        "External Key Management Service Service Account": ("gcp-sa-ekms.iam.gserviceaccount.com", None),
        "Firebase Extensions Service Agent": (
            "gcp-sa-firebasemods.iam.gserviceaccount.com",
            "roles/firebasemods.serviceAgent",
        ),
        "Firebase Rules Service Agent": ("firebase-rules.iam.gserviceaccount.com", None),
        "Firewall Insights Service Account": (
            "gcp-sa-firewallinsights.iam.gserviceaccount.com",
            "roles/firewallinsights.serviceAgent",
        ),
        "GKE Hub API Service Account": ("gcp-sa-gkehub.iam.gserviceaccount.com", "roles/gkehub.serviceAgent"),
        "Game Services Agent": ("gcp-sa-gameservices.iam.gserviceaccount.com", "roles/gameservices.serviceAgent"),
        "Google Cloud Dataproc Service Agent": (
            "dataproc-accounts.iam.gserviceaccount.com",
            "roles/dataproc.serviceAgent",
        ),
        "Google Cloud Functions Service Agent": (
            "gcf-admin-robot.iam.gserviceaccount.com",
            "roles/cloudfunctions.serviceAgent",
        ),
        "Google Cloud ML Engine Service Agent": (
            "cloud-ml.google.com.iam.gserviceaccount.com",
            "roles/ml.serviceAgent",
        ),
        "Google Cloud OS Config Service Agent": (
            "gcp-sa-osconfig.iam.gserviceaccount.com",
            "roles/osconfig.serviceAgent",
        ),
        "Google Cloud Run Service Agent": ("serverless-robot-prod.iam.gserviceaccount.com", "roles/run.serviceAgent"),
        "Google Container Registry Service Agent": (
            "containerregistry.iam.gserviceaccount.com",
            "roles/containerregistry.ServiceAgent",
        ),
        "Google Genomics Service Agent": (
            "genomics-api.google.com.iam.gserviceaccount.com",
            "roles/genomics.serviceAgent",
        ),
        "Kubernetes Engine Service Agent": (
            "container-engine-robot.iam.gserviceaccount.com",
            "roles/container.serviceAgent",
        ),
        "Mesh Config Service Account": ("gcp-sa-meshconfig.iam.gserviceaccount.com", "roles/meshconfig.serviceAgent"),
        "Monitoring Notification Service Account": (
            "gcp-sa-monitoring-notification.iam.gserviceaccount.com",
            "roles/monitoring.notificationServiceAgent",
        ),
        "Multi Cluster Ingress Service Account": (
            "gcp-sa-multiclusteringress.iam.gserviceaccount.com",
            "roles/multiclusteringress.serviceAgent",
        ),
        "Multi cluster metering Service Account": (
            "gcp-sa-mcmetering.iam.gserviceaccount.com",
            "roles/multiclustermetering.serviceAgent",
        ),
        "Private CA Service Account": ("gcp-sa-privateca.iam.gserviceaccount.com", None),
        "Remote Build Execution Service Agent": (
            "remotebuildexecution.iam.gserviceaccount.com",
            "roles/remotebuildexecution.serviceAgent",
        ),
        "Retail Service Account": ("gcp-sa-retail.iam.gserviceaccount.com", "roles/retail.serviceAgent"),
        "Secret Manager Service Account": ("gcp-sa-secretmanager.iam.gserviceaccount.com", None),
        "Serverless VPC Access Service Agent": (
            "gcp-sa-vpcaccess.iam.gserviceaccount.com",
            "roles/vpcaccess.serviceAgent",
        ),
        "Service Consumer Management Service Agent": ("service-consumer-management.iam.gserviceaccount.com", None),
        "Service Directory Service Account": ("gcp-sa-servicedirectory.iam.gserviceaccount.com", None),
        "Service Networking Service Agent": (
            "service-networking.iam.gserviceaccount.com",
            "roles/servicenetworking.serviceAgent",
        ),
        "TPU Service Agent": ("cloud-tpu.iam.gserviceaccount.com", "roles/tpu.serviceAgent"),
        "TPU Service Agent (v2)": ("gcp-sa-tpu.iam.gserviceaccount.com", "roles/cloudtpu.serviceAgent"),
        "Transcoder Service Account": ("gcp-sa-transcoder.iam.gserviceaccount.com", None),
    }

    _suffixes = ["Service Agent", "Agent", "Service Account"]

    @classmethod
    def get_available_agents(cls) -> Generator[str, None, None]:
        for agent in cls.agents:
            name = agent
            for suffix in cls._suffixes:
                name = name.replace(suffix, "")
            yield name

    @classmethod
    def _load_from_tsv(cls, filepath: str):
        # Load this <https://cloud.google.com/iam/docs/service-agents> table
        data = {}
        with Path(filepath).open(encoding="utf-8") as file:
            for line in file.readlines()[1:]:
                name, domain, role = line.strip().split("\t")
                role_name = None if "roles/" not in role else role.replace("(", "").replace(")", "")
                data[name] = (domain, role_name)
        return data

    @classmethod
    def _find(cls, service_name: str) -> tuple[str, str]:
        candidates = [f"{service_name} {suffix}" for suffix in cls._suffixes]

        for candidate in candidates:
            try:
                return cls.agents[candidate]
            except KeyError:
                continue

        raise exceptions.NotFound(f"Service {service_name} not found")

    @classmethod
    def get_email(cls, service_name: str, project_id: str) -> str:
        domain, _ = cls._find(service_name=service_name)
        project_number = cls.get_project_number(project_id=project_id)
        return f"service-{project_number}@{domain}"

    @classmethod
    def get_role(cls, service_name: str) -> str:
        _, role = cls._find(service_name=service_name)
        return role

    @classmethod
    def get_project_number(cls, project_id: str) -> int:
        # TODO: cache this
        project = ResourceManager().get_project(project_id=project_id)
        return project["projectNumber"]

    @classmethod
    def restore(cls, services: list[str], project_id: str) -> None:
        rm = ResourceManager()
        for service_name in services:
            email = ServiceAgent.get_email(service_name=service_name, project_id=project_id)
            role = ServiceAgent.get_role(service_name=service_name)
            if role:
                try:
                    rm.add_member(email=email, role=role)
                    logger.info(f"[O] {service_name}")
                except exceptions.ValidationError:
                    logger.warning(f"[X] {service_name}")

    @classmethod
    def get_compute_service_account(cls, project_id: str) -> str:
        project_number = cls.get_project_number(project_id=project_id)
        return f"{project_number}-compute@developer.gserviceaccount.com"

    @classmethod
    def get_cloud_build_service_account(cls, project_id: str | None = None) -> str:
        project_number = cls.get_project_number(project_id=project_id)
        return f"{project_number}@cloudbuild.gserviceaccount.com"


__all__ = (
    "ResourceManager",
    "ServiceAgent",
)
